//---------------------------------------------
//  Crest (Peak) Administration Application
//  (c) TASC Systems Inc. 2014-2016

//---------------------------------------------
// Peak Constants
//#region Peak Constants
var PEAK_WEB_SERVER_PORT = 80;
var WAIT_UNTIL_SYSTEM_STARTS = 500

var SnmpHostAddress = "10.10.7.39"
var SnmpCommunity = "public"

var ERRORLOG_FAILURE = 'FAILURE';
var ERRORLOG_WARNING = 'WARNING';
var ERRORLOG_NORMAL = 'NORMAL';
//#endregion Peak Constants

//---------------------------------------------
// Initializing Logging infrastructure
//#region Logging Infrastructure
// Start the logger
var log4js = require('log4js');

//config
log4js.configure({
    appenders: [
//        { type: 'console' }, // Comment this out for release
        {
            type: "file", 
            absolute: true, 
            filename: "./log/peak.log", 
            maxLogSize: 1000000,
            backups: 10,
            category: "crest"
        }
    ]
});
var logger = log4js.getLogger("crest");
logger.setLevel(log4js.levels.INFO);// Toggle to this if you want just basic log info *** for release ***
//logger.setLevel(log4js.levels.ALL);// Toggle to this if you want all log info *** for debug ***
//#endregion Logging Infrastructure

//---------------------------------------------
// Initializing server infrastructure
//#region Server Infrastructure
logger.trace('Intializing server infrastructure')

// Fire up the server + socket connections
var express = require('express');
var app = express();

// For serving static pages
var serveStatic = require('serve-static');

// For parsing middleware - e.g., JSON body
var bodyparser = require('body-parser');
app.use(bodyparser.json());
app.use(bodyparser.urlencoded({ extended: true }));

// For socket.io messaging and filesystem
var server = require('http').Server(app);
var io = require('socket.io')(server);
var fs = require('fs');
var path = require('path');
var moment = require('moment');
var os = require('os');
var pjson = require('./package.json')

logger.info('============================================================================================');
logger.info(pjson.description);
logger.info('   Version ' + pjson.version + ' (' + pjson.date + ') ')
logger.info('   (c) ' + pjson.author);
//logger.info('   Serial Number [coming soon: serial number of this device]');
logger.info('   ' + os.platform() + " OS Version " + os.release())
logger.info('   ' + os.freemem() + ' Free / ' + os.totalmem() + ' Total Memory');
logger.info('--------------------------------------------------------------------------------------------');

var ClientConnectedCounter = 0;
io.on('connect', function () {
    ClientConnectedCounter++;
    logger.info('Number of clients connected=' + ClientConnectedCounter);
});
io.on('disconnect', function () {
    ClientConnectedCounter--;
    logger.info('Number of clients connected=' + ClientConnectedCounter);
});
//#endregion Server Infrastructure

//---------------------------------------------
// Initializing Peak Users
//#region Peak Users
logger.info('Peak System: Intializing Peak User Management');
var peakUsers = require('./users.js');
peakUsers.Initialize(logger);
//#endregion Peak Users

//---------------------------------------------
// Initializing database infrastucture
//#region Database Infrastructure
logger.info('Peak System: Intializing Peak Database Object')
var sqlite3 = require('sqlite3').verbose();
var db = new sqlite3.Database('data/peak.db');
//#endregion Database Infrastructure

//---------------------------------------------
// Initialize System Status Manager
//#region System Status Manager
logger.info('Peak System: Intializing Peak System Monitor')
var systemStatusManager = require('./systemstatus.js')
systemStatusManager.InitializeSystemStatusManager(db, logger, function () {
    systemStatusManager.LogSystemStatus(pjson.description + " Version " + pjson.version + " started", ERRORLOG_NORMAL , fs, io, logger, db);
});
//#endregion System Status Manager

//---------------------------------------------
// Initializing network management
//#region Network Management
logger.info('Peak System: Intializing Peak Network, SMTP and SNMP objects')
var networkManager = require('./networkmanager.js')
var netSnmp = require('net-snmp');
var nodemailer = require('nodemailer');// SMTP support
networkManager.InitializeNetworkInterfaces(db, logger, netSnmp);

var smtpManager = require('./smtp.js');
smtpManager.InitializeSmtp(nodemailer, logger, function (smtp) {
    // Do nothing for now ... maybe later, maybe never
});
//#endregion Network Management

//---------------------------------------------
// Initializing Factory Reset
//#region Factory Reset
logger.info('Peak System: Intializing Factory Reset')
var factoryReset = require('./factoryreset.js');
//#endregion Factory Reset

//---------------------------------------------
// Initializing gps interface
//#region GPS Interface
var gpsInterface = require('./gps.js');

//#endregion GPS Interface

//---------------------------------------------
// Initializing audio interface
//#region Audio Manager
var audioManager = require('./audio.js');
var AudioClients = [];// Clients that might be listening in

//#endregion Audio Manager

//---------------------------------------------
// System Core
//#region System Core ==================================================================

//---------------------------------------------
//  Load PortsManager: ports and devices
//#region PortsManager start
logger.info("Peak System Core starting up!")
var portsManager = require('./ports.js');
logger.info('Peak System: Ports Manager - Intializing Ports Management Core')
portsManager.Initialize(logger, io, smtpManager, netSnmp, function () {
    portsManager.LoadDeviceModelsArray(function () {
        logger.info("Peak System: Ports Manager - Device models loaded")
        
        portsManager.LoadPortSettingsArray(function () {
            logger.info("Peak System: Ports Manager - Port settings loaded");
        });
    });
});
//#endregion PortsManager start

//---------------------------------------------
//  Load SNMP Hosts to each port/device
//#region SNMP Host setup
var SnmpHostsArray = [];// Global array of SNMP Hosts
var SnmpHostSessionsArray = [];// Global array of SNMP Host Sessions
networkManager.InitializeSnmpHostSettings(function (InitializedSnmpHostsArray) {
    // (Re)set the SNMP Host Array
    SnmpHostsArray = InitializedSnmpHostsArray.slice();
    logger.info("System Core: Initialized SnmpHostsArray=" + JSON.stringify(SnmpHostsArray));
    
    networkManager.CreateSnmpHostSessions(SnmpHostsArray, function (snmpHostSessions) {
        SnmpHostSessionsArray = snmpHostSessions.slice();
        logger.info("Peak System: Ports Manager - SNMP Hosts sessions created=" + SnmpHostSessionsArray.length)
        
        // Add delay to allow ports to load ... then set up SNMP sessions for each device
        setTimeout(function () {
            portsManager.SetSnmpHostSessionsArray(SnmpHostSessionsArray, function () {
                logger.info("Peak System: Ports Manager - SNMP Host sessions attached to Port Manager");
            });
        }, WAIT_UNTIL_SYSTEM_STARTS);
    });
});
//#endregion SNMP Host setup

//---------------------------------------------
//  Factory reset processing starts here....
//#region Factory Reset
logger.info("System Core: Factory Reset Processing starting")
factoryReset.ProcessFactoryReset(db, logger, io, netSnmp, systemStatusManager, function () { });
//#endregion Factory Reset

//#endregion System Core ==================================================================

//---------------------------------------------
// Initializing API middleware
//#region API Middleware
logger.info('Peak System: Initializing Crest API')
app.use(serveStatic('.', { 'index': ['index.html'] }))
app.use(serveStatic('views', { 'index': false }))
//#endregion API Middleware 

//---------------------------------------------
// Utility functions
//#region Utility functions
app.get('/api/causeerror', function (req, res, next) {
    logger.info('Crest API: GET /api/causeerror');
    process.exit(1);
    logger.info('changed this file');
    res.send('caused error');
})
//#endregion Utility functions

//#region Peak User

// Login Peak User
app.put('/api/login', function (req, res, next) {
    logger.info('Crest API: PUT /api/login');
    var ip = req.headers['x-real-ip'] || req.connection.remoteAddress
    peakUsers.LoginUser(req.body, ip, function (success) {
        if (success) {
            logger.info('Crest API: Login attempt by user=' + req.body.user + " was successful");
            res.send({ "LoginPassed": true });
        }        
        else {
            logger.info('Crest API: Login attempt by user=' + req.body.user + " failed");
            res.send({ "LoginPassed": false });
        }
    });
});

// Get Peak Login Attempts
app.get('/api/login/attempts', function (req, res, next) {
    logger.info('Crest API: GET /api/login/attempts');
    peakUsers.GetPeakLoginAttempts(function (PeakLoginAttempts) {
        logger.info("Crest API: Peak Login Attempts=" + PeakLoginAttempts.length);
        res.send(PeakLoginAttempts);
    });
});

// Get Peak User
app.get('/api/login/users', function (req, res, next) {
    logger.info('Crest API: GET /api/login/users');
    peakUsers.GetPeakUser(function (PeakUser) {
        logger.info("Crest API: Peak User=" + PeakUser.user);
        res.send(PeakUser);
    });
});

// Put Peak User
app.put('/api/login/users', function (req, res, next) {
    logger.info('Crest API: PUT /api/login/users');
    peakUsers.PutPeakUser(req.body, function (PeakUser) {
        logger.info("Crest API: Peak User=" + PeakUser.user);
        res.send(PeakUser);
    });
});

//#endregion Peak User

//---------------------------------------------
// Peak System
//#region Peak System

// Identify system
app.get('/api/peak/identify', function (req, res, next) {
    logger.info('Crest API: GET /api/peak/identify');
    // Flash lights yellow 1s, then blue 1s for a total of 1 minute
    var timeDelay = 3000;
    for (var i = 0; i < 5; i++) {
        setTimeout(function () {
            systemStatusManager.WarningOperation(fs, io, logger);
        }, timeDelay * i);
        
        setTimeout(function () {
            systemStatusManager.NormalOperation(fs, io, logger);
        }, timeDelay * i + (timeDelay / 2));
    }
    res.send('Identify Device OK');
})

// Get installed system software
app.get('/api/peak/systemsoftware', function (req, res, next) {
    logger.info('Crest API: GET /api/peak/systemsoftware');
    logger.trace('Crest API: GET /api/peak/systemsoftware - System Software=' + JSON.stringify(pjson))
    res.send(pjson);
})

// Get installed system software
app.get('/api/peak/operatingsystem', function (req, res, next) {
    logger.info('Crest API: GET /api/peak/operatingsystem');
    data = { "os": os.platform() + " OS Version " + os.release() };
    logger.trace('Crest API: GET /api/peak/operatingsystem - System OS=' + JSON.stringify(data))
    res.send(data);
})

app.get('/api/peak/latestsystemsoftware', function (req, res, next) {
    logger.info('Crest API: GET /api/peak/latestsystemsoftware');
    systemStatusManager.GetLatestAppVersionFromServer(logger, function (latestversion) {
        logger.trace('Crest API: GET /api/peak/latestsystemsoftware - Latest Version=' + JSON.stringify(latestversion))
        res.send(latestversion);
    });
})

// Get current system status
app.get('/api/peak/systemstatus', function (req, res, next) {
    logger.info('Crest API: GET /api/peak/systemstatus');
    var CurrentSystemStatus = systemStatusManager.GetCurrentSystemStatus(logger);
    logger.trace('Crest API: GET /api/peak/systemstatus - Current System Status=' + CurrentSystemStatus);
    res.send(CurrentSystemStatus);
})

// Get list of system events
app.get('/api/peak/systemevents', function (req, res, next) {
    logger.info('Crest API: GET /api/peak/systemstatus');
    var systemStatus = [];
    db.all("SELECT ErrorMessage, ErrorType, DATE(TimeChanged) FROM ErrorLog ORDER BY DATE(TimeChanged) DESC LIMIT 20", function (err, rows) {
        if (err) {
            throw logger.error("Crest API: GET /api/peak/systemstatus - ERROR=" + err);
        }
        for (i = 0; i < rows.length; i++) {
            // Always send the raw value - clients responsible for conversion
            systemStatus.push({
                "ErrorType"     : rows[i].ErrorType,
                "ErrorMessage"  : rows[i].ErrorMessage,
                "TimeChanged"   : rows[i].TimeChanged
            });
        }
        logger.trace('Crest API: GET /api/peak/systemstatus - System Status=' + JSON.stringify(systemStatus));
        res.json(systemStatus);
    });
})
//#endregion

//---------------------------------------------
// Crest Software
//#region Crest Software
// Get Crest software info
app.get('/api/crest', function (req, res, next) {
    logger.info('Crest API: GET /api/crest');
    logger.trace('Crest API: GET /api/crest - Crest info=' + JSON.stringify(pjson))
    res.send(pjson);
})

app.get('/api/crest/latest', function (req, res, next) {
    logger.info('Crest API: GET /api/crest/latest');
    systemStatusManager.GetLatestCrestVersionFromServer(logger, function (latestversion) {
        logger.trace('Crest API: GET /api/crest/latest - Crest latest version=' + latestversion)
        res.send(latestversion);
    });
})

app.put('/api/crest/download', function (req, res, next) {
    logger.info('Crest API: GET /api/crest/download');
    logger.info('Crest API: GET /api/crest/download ' + JSON.stringify(req.body));
    systemStatusManager.DownloadCrest(req.body.CrestUrl, logger, function (result) {
        logger.trace('Crest API: GET /api/crest/download - Crest download=' + result);
        res.send(result);
    });
})

//#endregion Crest Software

//#region Networking, SMTP and SNMP API
//---------------------------------------------
// Networking, SMTP and SNMP

// Networking Support
//#region Networking Support
// Get network interfaces
app.get('/api/network/interfaces', function (req, res, next) {
    
    // Get Network Interfaces information from the Network Manager module
    logger.info('Crest API: GET /api/network/interfaces - Get Network Interfaces called');
    networkManager.GetNetworkInterfaces(function (err, networkInterfaces) {
        if (err) {
            logger.info('Crest API: GET /api/network/interfaces - Return Network Interfaces. Error=' + err);
            res.json({});
        } else {
            // Return the network interfaces
            logger.info('Crest API: GET /api/network/interfaces - Return Network Interfaces=' + JSON.stringify(networkInterfaces));
            res.json(networkInterfaces);
        }
    });

})

// Set network interfaces
app.put('/api/network/interfaces', function (req, res, next) {
    logger.info('Crest API: PUT /api/network/interfaces - Updating Network Interface=' + JSON.stringify(req.body));
    networkManager.PutNetworkInterfaces(req.body, function (error, interfaces) {
        if (error) {
            logger.error('Crest API: PUT /api/network/interfaces - Updated Network Interface. Error=' + error);
        } else {
            logger.info('Crest API: PUT /api/network/interfaces - Updated Network Interface=' + JSON.stringify(interfaces));
        }
        res.send(interfaces);
    });
})
//#endregion Networking Support

// SMTP Support
//#region SMTP Support

// Get SMTP Settings
app.get('/api/network/smtp', function (req, res, next) {
    logger.info('Crest API: GET /api/network/smtp');
    smtpManager.GetSmtpSettings(function (smtp) {
        logger.info('Crest API: GET /api/network/smtp - SMTP Settings=' + smtp);
        res.json(smtp);
    })
});

// Put SMTP Settings
app.put('/api/network/smtp', function (req, res, next) {
    logger.info('Crest API: PUT /api/network/smtp');
    req.body.Default = false;// No longer using default value
    smtpManager.SetSmtpSettings(req.body, function (smtp) {
        logger.info('Crest API: PUT /api/network/smtp - SMTP Settings=' + smtp);
        res.json(smtp);
    })
});

// Test a SMTP host 
app.put('/api/network/smtp/test', function (req, res, next) {
    var password = req.body.Pass;
    req.body.Pass = "HIDDEN";
    logger.info('Crest API: PUT /api/network/smtp/test - data=' + JSON.stringify(req.body));
    req.body.Pass = password;
    
    // create reusable transporter object using the default SMTP transport
    var transporter = nodemailer.createTransport({
        port: req.body.Port,
        host: req.body.Host,
        secure: req.body.Secure,
        auth: {
            user: req.body.User,
            pass: req.body.Pass
        }
    });
    
    // setup e-mail data with unicode symbols
    var mailOptions = {
        from: req.body.From, // sender address
        to: req.body.To, // list of receivers
        subject: 'Crest SMTP Test Email', // Subject line
        text: 'Your SMTP configuration is OK and so Crest is able to send you emails!', // plaintext body
    };
    
    // send mail with defined transport object
    var result = false;
    transporter.sendMail(mailOptions, function (error, info) {
        if (error) {
            logger.error("Crest API: PUT /api/network/smtp/test - send email error=" + error);
            result = false;
            res.json({ "result": result , "response": error });
        } else {
            logger.info("Crest API: PUT /api/network/smtp/test - send email result=" + info.response);
            result = true;
            res.json({ "result": result, "response": info.response });
        }
    });
});

//#endregion SMTP Support

// SNMP Support
//#region SNMP Support
// Get SNMP hosts
app.get('/api/network/snmp/hosts', function (req, res, next) {
    logger.info('Crest API: GET /api/network/snmp/hosts');
    
    // Get the complete list of SNMP Hosts and update I/O processing sub-systems
    var snmpHosts = [];
    db.all("SELECT rowid as Id, HostName, Enabled, IpAddress, Port, Community FROM snmphosts ORDER BY Id", function (err, rows) {
        if (err) {
            // LogSystemStatus(errorMessage, errorType, timeStamp, db, systemStatusManager)
            throw logger.error('Crest API: GET /api/network/snmp/hosts - Error=' + err);
        }
        for (i = 0; i < rows.length; i++) {
            // Always send the raw value - clients responsible for conversion
            snmpHosts.push({
                "Id"            : rows[i].Id,
                "HostName"      : rows[i].HostName,
                "Enabled"       : rows[i].Enabled,
                "IpAddress"     : rows[i].IpAddress,
                "Port"          : rows[i].Port,
                "Community"     : rows[i].Community
            });
        }
        logger.debug('Crest API: GET /api/network/snmp/hosts - Hosts=' + JSON.stringify(snmpHosts));
        res.json(snmpHosts);
    });
})

// Delete a SNMP host
app.put('/api/network/snmp/hosts/delete', function (req, res, next) {
    //  Id          (TEXT)      <Id of the host>
    //  HostName    (TEXT)      <host name>
    //  Enabled     (TEXT)      ”TRUE”,”FALSE”
    //  IpAddress   (TEXT)      <IP Address of Host>
    //  Port        (TEXT)      <Port for host, typically 162>
    //  Community   (TEXT)      <SNMP community name>
    logger.info('Crest API: PUT /api/network/snmp/hosts/delete');
    
    // Update the existing record
    logger.debug('Crest API: PUT /api/network/snmp/hosts/delete - Host to delete=' + JSON.stringify(req.body));
    
    var update = "DELETE FROM SnmpHosts WHERE rowid='" + req.body.Id + "'";
    db.run(update, function (err) {
        if (err !== null) {
            logger.error('Crest API: PUT /api/network/snmp/hosts/delete - Error (Update)=' + JSON.stringify(err));
            res.send({ msg: "SQLite3 dB: api/network/snmp/hosts/" + " DELETE FROM  err=" + JSON.stringify(err) })
        }
        
        // Get the complete list of SNMP Hosts and update I/O processing sub-systems
        var snmpHosts = [];
        db.all("SELECT rowid as Id, HostName, Enabled, IpAddress, Port, Community FROM snmphosts ORDER BY Id", function (err, rows) {
            if (err) {
                throw logger.error('Crest API: PUT /api/network/snmp/hosts/delete - Error (Read)=' + JSON.stringify(err));
            }
            for (i = 0; i < rows.length; i++) {
                // Always send the raw value - clients responsible for conversion
                snmpHosts.push({
                    "Id"            : rows[i].Id,
                    "HostName"      : rows[i].HostName,
                    "Enabled"       : rows[i].Enabled,
                    "IpAddress"     : rows[i].IpAddress,
                    "Port"          : rows[i].Port,
                    "Community"     : rows[i].Community
                });
            }
            logger.debug('Crest API: PUT /api/network/snmp/hosts/delete - Hosts=' + JSON.stringify(snmpHosts));
            res.json(snmpHosts);
        });
    });
})

// Create new SNMP host
app.put('/api/network/snmp/hosts', function (req, res, next) {
    //  HostName    (TEXT)      <host name>
    //  Enabled     (TEXT)      ”TRUE”,”FALSE”
    //  IpAddress   (TEXT)      <IP Address of Host>
    //  Port        (TEXT)      <Port for host, typically 162>
    //  Community   (TEXT)      <SNMP community name>
    logger.info('Crest API: PUT /api/network/snmp/hosts');
    
    // Add a new record
    logger.debug('Crest API: PUT /api/network/snmp/hosts - Host to add=' + JSON.stringify(req.body));
    
    // Insert the new record
    db.run("BEGIN")
    insertStatement = "'" + req.body.HostName + "', '" + req.body.Enabled + "', '" + req.body.IpAddress + "', '" + req.body.Port + "', '" + req.body.Community + "'";
    db.run("INSERT INTO SnmpHosts (HostName, Enabled, IpAddress, Port, Community) VALUES (" + insertStatement + ")", function (err) {
        if (err !== null) {
            logger.error('Crest API: PUT /api/network/snmp/hosts - Error (Insert)=' + err);
            res.send({ msg: "SQLite3 dB: api/snmphosts Insert err=" + JSON.stringify(err) })
        }
        
        // Successful insert, just look up the ID for future reference
        logger.debug('Crest API: PUT /api/network/snmp/hosts - ID=' + this.lastID);
        req.body.Id = this.lastID;
    });
    db.run("COMMIT");
    
    // Get the complete list of SNMP Hosts and update I/O processing sub-systems
    var snmpHosts = [];
    db.all("SELECT rowid as Id, HostName, Enabled, IpAddress, Port, Community FROM snmphosts ORDER BY Id", function (err, rows) {
        if (err) {
            // LogSystemStatus(errorMessage, errorType, timeStamp, db, systemStatusManager)
            throw logger.error('Crest API: PUT /api/network/snmp/hosts - Error (Read)=' + err);
        }
        for (i = 0; i < rows.length; i++) {
            // Always send the raw value - clients responsible for conversion
            snmpHosts.push({
                "Id"            : rows[i].Id,
                "HostName"      : rows[i].HostName,
                "Enabled"       : rows[i].Enabled,
                "IpAddress"     : rows[i].IpAddress,
                "Port"          : rows[i].Port,
                "Community"     : rows[i].Community
            });
        }
        logger.debug('Crest API: PUT /api/network/snmp/hosts - Hosts=' + JSON.stringify(snmpHosts));
        
        // (Re)set the SNMP Host Sessions for I/O Processors
        networkManager.CreateSnmpHostSessions(snmpHosts, function (sessions) {
            portsManager.SetSnmpHostSessionsArray(sessions, function () {
                res.json(snmpHosts);
            });
        });
    });
})

// Modify existing SNMP host
app.put('/api/network/snmp/hosts/:snmphost', function (req, res, next) {
    //  Id          (TEXT)      <Id of the host>
    //  HostName    (TEXT)      <host name>
    //  Enabled     (TEXT)      ”TRUE”,”FALSE”
    //  IpAddress   (TEXT)      <IP Address of Host>
    //  Port        (TEXT)      <Port for host, typically 162>
    //  Community   (TEXT)      <SNMP community name>
    logger.info('Crest API: PUT /api/network/snmp/hosts/' + req.params.snmphost);
    
    // Update the existing record
    var update = "UPDATE SnmpHosts SET HostName='" + req.body.HostName + "', Enabled='" + req.body.Enabled + "', IpAddress='" + req.body.IpAddress + "', Port='" + req.body.Port + "', Community='" + req.body.Community + "' WHERE rowid='" + req.params.snmphost + "'";
    db.run(update, function (err) {
        if (err !== null) {
            throw logger.error('Crest API: PUT /api/network/snmp/hosts/' + req.params.snmphost + ' - Error (Update)=' + JSON.stringify(err));
            res.send({ msg: "SQLite3 dB: api/snmphosts/" + req.body.Id + " UPDATE err=" + JSON.stringify(err) })
        }
    });
    
    // Get the complete list of SNMP Hosts and update I/O processing sub-systems
    var snmpHosts = [];
    db.all("SELECT rowid as Id, HostName, Enabled, IpAddress, Port, Community FROM snmphosts ORDER BY Id", function (err, rows) {
        if (err) {
            throw logger.error('Crest API: PUT /api/network/snmp/hosts/' + req.params.snmphost + ' - Error (Read)=' + err);
        }
        for (i = 0; i < rows.length; i++) {
            // Always send the raw value - clients responsible for conversion
            snmpHosts.push({
                "Id"            : rows[i].Id,
                "HostName"      : rows[i].HostName,
                "Enabled"       : rows[i].Enabled,
                "IpAddress"     : rows[i].IpAddress,
                "Port"          : rows[i].Port,
                "Community"     : rows[i].Community
            });
            
            if (rows[i].Id == req.params.snmphost) {
                snmpHostReturn = rows[i];
            }
        }
        logger.trace('Crest API: PUT /api/network/snmp/hosts/' + req.params.snmphost + ' - Hosts=' + JSON.stringify(snmpHosts));
        
        // (Re)set the SNMP Host Sessions for I/O Processors
        networkManager.CreateSnmpHostSessions(snmpHosts, function (sessions) {
            portsManager.SetSnmpHostSessionsArray(sessions, function () {
                res.send(snmpHostReturn);
            });
        });
    });
})

// Test a SNMP host by sending a trap message
app.put('/api/network/snmp/test', function (req, res, next) {
    logger.info('Crest API: PUT /api/network/snmp/test');
    
    logger.trace('Crest API: PUT /api/network/snmp/test - Host=' + JSON.stringify(req.body));
    // Create session for test send
    var SnmpSessionOptions = {
        port: 161,
        retries: 1,
        timeout: 5000,
        transport: "udp4",
        trapPort: 162,
        version: netSnmp.Version2c
    };
    SnmpHostAddress = req.body.IpAddress;
    SnmpCommunity = req.body.Community;
    SnmpOptions = JSON.parse(JSON.stringify(SnmpSessionOptions));
    SnmpOptions.trapPort = req.body.Port;
    TestSession = netSnmp.createSession(SnmpHostAddress, SnmpCommunity, SnmpOptions);
    
    // Build the trap
    var trapOid = "1.3.6.1.4.1.11902.10.5.1";
    var varbinds = [
        {
            oid: "1.3.6.1.4.1.11902.10.5.2",
            type: netSnmp.ObjectType.OctetString,
            value: "Test Trap"
        },
    ];
    
    // Send the trap
    TestSession.trap(trapOid, varbinds, function (error) {
        if (error)
            logger.error('Crest API: PUT /api/network/snmp/test - Error (Trap send)=' + error);
        else
            logger.debug('Crest API: PUT /api/network/snmp/test - Success');
    });
    res.send(req.body);
});

// Return the TASC-PEAK.MIB file
app.get('/api/network/snmp/mibfile', function (req, res, next) {
    logger.info('Crest API: GET /api/network/snmp/mibfile');
    
    // Read the TASC-PEAK.MIB file
    fs.readFile("./TASC-PEAK.MIB", 'utf8', function (err, data) {
        if (err) {
            logger.error('Crest API: GET /api/network/snmp/mibfile - Error (Read)=' + JSON.stringify(err));
            res.send("Error reading TASC-PEAK.MIB file");
        }
        else {
            logger.debug('Crest API: GET /api/network/snmp/mibfile - File=TASC-PEAK.MIB');
            res.send(encodeURIComponent(data));
        }
    });
})
//#endregion SNMP Support

//#endregion Networking, SMTP and SNMP API

//---------------------------------------------
// Logging API
//#region Logging API

app.get('/api/logging/loglevel', function (req, res, next) {
    logger.info('Crest API: GET /api/logging/loglevel');
    
    var loglevel;
    switch (logger.level.levelStr) {
        case "INFO":
            loglevel = "BASIC";
            break;

        case "TRACE":
            loglevel = "VERBOSE";
            break;

        case "ALL":
            loglevel = "ALL";
            break;

    }
    logger.info('Crest API: PUT /api/logging/loglevel - Level=' + loglevel);
    res.json({ "LogLevel": loglevel });
});

app.put('/api/logging/loglevel/:loglevel', function (req, res, next) {
    logger.info('Crest API: PUT /api/logging/loglevel/' + req.params.loglevel);
    
    switch (req.params.loglevel) {
        case "BASIC":
            logger.setLevel(log4js.levels.INFO);
            loglevel = "BASIC";
            logger.info("Logging Manager - Level=" + loglevel + " (" + logger.level.levelStr + ")");
            break;

        case "VERBOSE":
            logger.setLevel(log4js.levels.TRACE);
            loglevel = "VERBOSE";
            logger.info("Logging Manager - Level=" + loglevel + " (" + logger.level.levelStr + ")");
            break;

        case "ALL":
            logger.setLevel(log4js.levels.ALL);
            loglevel = "ALL";
            logger.info("Logging Manager - Level=" + loglevel + " (" + logger.level.levelStr + ")");
            break;
    }
    logger.info('Crest API: PUT /api/logging/loglevel/' + req.params.loglevel + " - Level=" + logger.level.levelStr);
    res.json({ "LogLevel": loglevel });
});

// Get list of log files
app.get('/api/logging/logfiles', function (req, res, next) {
    logger.info('Crest API: GET /api/logging/logfiles');
    
    // Read the log files
    logDirectory = './log';
    var logFiles = [];
    fs.readdir(logDirectory, function (err, files) {
        if (err) {
            logger.error('Crest API: GET /api/logging/logfiles - Error (Read)=' + JSON.stringify(err));
            res.send('Error reading log files');
        }
        else {
            // Get a list of these files
            for (var i in files) {
                var filePath = logDirectory + '/' + files[i];
                var stat = fs.statSync(filePath);
                if (stat.isFile()) {
                    logFiles.push({
                        "Name"  : files[i],
                        "Size"  : parseInt(stat.size).toLocaleString(),
                        "Date"  : stat.mtime
                    });
                }
            }
            
            // Put the files in the order that we want
            logFiles.sort().reverse(); // this will put 'peak.log' at end, but sort with most recent at front 
            var peakLog = logFiles.pop();// get 'peak.log'
            logFiles.unshift(peakLog);// put 'peak.log' in front
            
            logger.trace('Crest API: GET /api/logging/logfiles - Files=' + JSON.stringify(logFiles));
            res.send(logFiles);
        }
    });
});

// Read contents of a log file
app.get('/api/logging/logfiles/:logfile', function (req, res, next) {
    logger.info('Crest API: GET /api/logging/logfiles/' + req.params.logfile);
    
    // Read the log files
    logDirectory = './log';
    fs.readFile(logDirectory + "/" + req.params.logfile, 'utf8', function (err, data) {
        if (err) {
            logger.error('Crest API: GET /api/logging/logfiles/' + req.params.logfile + ' - Error (Read)=' + JSON.stringify(err));
            res.send("Error reading log file");
        }
        else {
            logger.trace('Crest API: GET /api/logging/logfiles - File=' + req.params.logfile);
            res.send(encodeURIComponent(data));
        }
    });
})
//#endregion Logging API

//---------------------------------------------
// Audio
//#region Audio

app.get('/api/audio/stream/:control', function (req, res, next) {
    logger.info('Crest API: GET /api/audio/stream/:' + req.params.control);
    
    var result = "Error"
    if (req.params.control === "on") {
        // Start streaming audio
        audioManager.AudioStart({ 'mp3output' : true }, logger);
        result = "Audio Started";
    } else if (req.params.control === "off") {
        // Stop streaming audio
        audioManager.AudioStop(logger);
        result = "Audio Stopped";
    }
    
    res.json({ "Result": result });
});

// If there's someone there to receive it, 
//  send a chunk of what we received on the mic
audioManager.AudioStream.on('data', function (data) {
    if (AudioClients.length > 0) {
        for (client in AudioClients) {
            AudioClients[client].write(data);
        }        ;
        //logger.info("AudioManager - Sending mic data to clients: bytes=" + data.length);
    }
});

var audioServer = require('http');
audioServer.createServer(function (request, response) {
    logger.info('AudioManager - createServer');
    response.writeHead(200, {
        "Content-Type": "audio/mpeg"
        //,
        //'Transfer-Encoding': 'chunked'
    });
    
    // Add the response to the clients array to receive streaming
    AudioClients.push(response);
    logger.info('AudioManager - Audio client connected - total audio clients connected=' + AudioClients.length);
}).listen(2000);// Listen on a web port and respond with a chunked response header.

//#endregion Audio

//---------------------------------------------
// Ports
//#region Ports

// Get Ports
app.get('/api/ports', function (req, res, next) {
    logger.info('Crest API: GET /api/ports');
    
    var ports = portsManager.GetPortSettingsArray();
    logger.info('Crest API: GET /api/ports - Ports=' + JSON.stringify(ports));
    res.json(ports);
})

// Update Port
app.put('/api/ports/:port', function (req, res, next) {
    logger.info('Crest API: PUT /api/ports/' + req.params.port + ' device=' + JSON.stringify(req.body));

    portsManager.UpdatePortSettingsArrayPort(req.body, function (ports) {
        logger.info('Crest API: PUT /api/ports/' + req.params.port + ' Adjusted PortSettingsArray=' + JSON.stringify(ports));
        res.json(ports);
    });
})

//#endregion Ports

//---------------------------------------------
// Devices
//#region Devices

// Get DeviceGroups
app.get('/api/devicegroups', function (req, res, next) {
    logger.info('Crest API: GET /api/devicegroups');
    
    var deviceGroups = portsManager.GetDeviceGroupsArray();
    logger.trace('Crest API: GET /api/devicegroups - DeviceGroups=' + JSON.stringify(deviceGroups));
    res.json(deviceGroups);
})

// Get Devices
app.get('/api/devices', function (req, res, next) {
    logger.info('Crest API: GET /api/devices');
    
    var devices = portsManager.GetDeviceModelsArray();
    logger.info('Crest API: GET /api/devices - Devices=' + JSON.stringify(devices));
    res.json(devices);
})

// Get Device Info
app.get('/api/devices/info/:port', function (req, res, next) {
    logger.info('Crest API: GET /api/devices/info/' + req.params.port);

    portsManager.GetDeviceInformation(req.params.port, function (deviceInfo) {
        logger.info('Crest API: GET /api/devices/info/' + req.params.port + ' - Devices=' + JSON.stringify(deviceInfo));
        res.json(deviceInfo);
    });
})

// Update Device
app.put('/api/devices/:port', function (req, res, next) {
    logger.info('Crest API: PUT /api/devices/' + req.params.port + ' device=' + JSON.stringify(req.body));
    
    portsManager.UpdatePortSettingsArrayPortDevice(req.params.port, req.body, function (updatedmodel) {
        logger.info('Crest API: PUT /api/devices/' + req.params.port + ' device=' + JSON.stringify(updatedmodel));
        res.json(updatedmodel);
    });
})

//#endregion Devices

//---------------------------------------------
// EventLog
//#region EventLog

// Get EventLog
app.get('/api/eventlog/:port', function (req, res, next) {
    logger.info('Crest API: GET /api/eventlog/' + req.params.port);
    
    // Retrieve EventLog array from port x log file
    var EventLog = [];
    if (fs.existsSync("log/port" + req.params.port + "events.json")) {
        fs.readFile("log/port" + req.params.port + "events.json", function (error, eventlog) {
            if (error) {
                logger.error("Crest API: GET /api/eventlog/" + req.params.port + ": EventLog retrieve error. Error=" + error);
                res.json(EventLog);
            } else {
                EventLog = JSON.parse(eventlog);
                logger.trace("Crest API: GET /api/eventlog/" + req.params.port + ": EventLog=" + JSON.stringify(EventLog));
                res.json(EventLog);
            }
        });
    } else {
        logger.warn("Crest API: GET /api/eventlog/" + req.params.port + ": log/port" + req.params.port + "events.json does not exist");
        res.json(EventLog);
    }
})

//#endregion EventLog

//---------------------------------------------
// System Clock 
//#region System Clock
logger.info('Intializing System Clock')
setInterval(sendTime, 1000)
function sendTime() {
    var currentTime = moment();
    io.sockets.emit("CurrentSystemTime", { "CurrentSystemTime": currentTime });
}
//#endregion System Clock

//---------------------------------------------
// Initialize Client Connections
//#region Client Connections
server.listen(PEAK_WEB_SERVER_PORT, function () {
    logger.info('Web server listening on port ' + PEAK_WEB_SERVER_PORT)
});

io.on('connection', function (socket) {
    logger.info('Real-time reporting via socket connection is established');
});
//#endregion Client Connections